"""
Contains patches that modify the binary at the instruction level.
"""

from __future__ import annotations

import logging
from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING

from ..components.allocation_managers.allocation_manager import MemoryFlag
from .patch import Patch

logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from ..patcherex import Patcherex


class ModifyInstructionPatch(Patch):
    """
    Patch that directly modifies instructions in a binary (overwrites them) starting at address given.
    If ISA is variable length, then if there are remaining bytes in the last overwritten instruction,
    it will fill them with nops, but it will fail if remaining bytes are not divisible by nop length.
    """

    def __init__(
        self, addr: int, instr: str, symbols: dict[str, int] | None = None
    ) -> None:
        """
        Constructor.

        :param addr: Memory address of instruction(s) to overwrite.
        :param instr: Assembly instruction(s) to place in binary. If you want to use any symbols from the program or from previous patches, you must surround them with curly braces.
        :param symbols: Symbols to include when assembling, in format {symbol name: memory address}, defaults to None
        """
        self.addr = addr
        self.instr = instr
        self.symbols = symbols if symbols else {}

    def apply(self, p) -> None:
        """
        Applies the patch to the binary, intended to be called by a Patcherex instance.

        :param p: Patcherex instance.
        :type p: Patcherex
        """
        asm_bytes = p.assembler.assemble(
            self.instr,
            self.addr,
            symbols=self.symbols,
            is_thumb=p.binary_analyzer.is_thumb(self.addr),
        )
        if p.archinfo.is_variable_length_isa:
            asm_size = len(asm_bytes)
            overwritten_size = 0
            num_instrs = 1
            while overwritten_size < asm_size:
                overwritten_size = len(
                    p.binary_analyzer.get_instr_bytes_at(
                        self.addr, num_instr=num_instrs
                    )
                )
                num_instrs += 1
            remaining_size = overwritten_size - asm_size
            assert remaining_size % p.archinfo.nop_size == 0, (
                f"Cannot fill in {remaining_size} bytes when modifying instruction, must be a multiple of {p.archinfo.nop_size}"
            )
            asm_bytes += p.archinfo.nop_bytes * (remaining_size // p.archinfo.nop_size)
        offset = p.binary_analyzer.mem_addr_to_file_offset(self.addr)
        p.binfmt_tool.update_binary_content(offset, asm_bytes)


def convert_to_subregisters(
    cc: list[str],
    subregisters: dict[str, dict[int, list[str]]],
    regs_sort: list[str | tuple[str, int] | tuple[str, str]],
    type_converter: Callable[[int], str],
) -> list[tuple[str, str]]:
    regs = [r if isinstance(r, str) else r[0] for r in regs_sort]

    # reg_sizes maps registers to their size
    reg_sizes: dict[str, int] = dict()
    # parent_regs maps all children registers to their largest parent register
    parent_regs: dict[str, str] = dict()
    for parent, subregister_info in subregisters.items():
        for child_bits, children in subregister_info.items():
            for i, child in enumerate(children):
                if i > 0 and child in regs:
                    # Only allow the 0th subregister to actually be used.
                    raise ValueError(
                        f"Unable to create the calling convention when the {child} register is present. The {children[0]} subregister is the only {child_bits} bit subregister that can be used."
                    )
                reg_sizes[child] = child_bits
                parent_regs[child] = parent

    # The rewrites that should be applied to the cc to compute the transformed cc output
    rewrites: dict[str, tuple[str, str]] = dict()
    for r in regs_sort:
        if isinstance(r, str):
            reg_name = r
        else:
            reg_name = r[0]
        if reg_name in parent_regs:
            if isinstance(r, str):
                reg_bits = reg_sizes[reg_name]
                reg_type = type_converter(reg_bits)
            elif isinstance(r[1], int):
                (_, reg_bits) = r
                reg_type = type_converter(reg_bits)
            else:
                (_, reg_type) = r
            parent = parent_regs[reg_name]
            if parent in rewrites:
                raise ValueError(
                    "The following two input registers overlapped while computing the calling convention: "
                    + r
                    + " and "
                    + rewrites[parent][1]
                )
            rewrites[parent] = (reg_name, reg_type)

    def convert_cc_reg(cc_reg):
        if cc_reg in rewrites:
            return rewrites[cc_reg]
        else:
            parent_reg = parent_regs[cc_reg]
            parent_bits = reg_sizes[parent_reg]
            parent_type = type_converter(parent_bits)
            return (parent_reg, parent_type)

    return list(map(convert_cc_reg, cc))


class InsertInstructionPatch(Patch):
    class CConfig:
        def __init__(
            self,
            c_forward_header: str = "",
            scratch_regs: Iterable[str] = None,
            regs_sort: Iterable[str | tuple[str, int] | tuple[str, str]] | None = None,
            asm_header: str = "",
            asm_footer: str = "",
        ):
            """
            Used to configure an InsertInstructionPatch when language == "C"

            :param c_forward_header: C code that will be inserted before the micropatch code. This is useful when you want\
            to use C types, C headers, and C function forward declarations.
            :param scratch_regs: It is generally a good idea to mark some registers as scratch to give the compiler\
            breathing room for allocating registers to use for intermediate variables in your micropatch.\
            All of the registers that we mark as scratch can be freely clobbered by the compiler\
            Note that you can still read from scratch registers stored in the variables. What the scratch\
            register denotation will indicate however is that the register can be re-used after the variable\
            is no longer live.
            :param regs_sort: Some architectures have subregisters\
            which allow access to a portion of a register, such as the lower 32 bits of a 64 bit register. If you want\
            to use a subregister instead of the full register you can request this by passing in a list of subregisters\
            here. Note that if you specify a subregister here, the full parent register is not available in the C\
            patch. Some registers allow values of smaller types to be stored in larger registers. For example, you can\
            store float values in xmm0 on x64, but there is no 32 bit subregister. To indicate that a register\
            should be treated as a specific size, you may specify register size in bits in the second element of the\
            tuple. Alternatively, you may specify the type as a string. An example of a valid list for x64 is\
            ['eax', ('xmm0', 32), ('xmm1', 'double')].
            :param asm_header: The asm_header is inserted in the main body of the patch before the C code. This header is\
            primarily useful for gaining access to the stack pointer, which is a register that is typically unavailable in\
            our C patch code.
            :param asm_footer: The asm_footer is the same as the asm_header, except that it runs after the C code executes.\
            This is typically less useful than asm_header, but is still available here if you need to do any cleanup in\
            assembly.
            """
            self.c_forward_header = c_forward_header
            self.scratch_regs = scratch_regs
            self.regs_sort = regs_sort
            self.asm_header = asm_header
            self.asm_footer = asm_footer

    """
    Patch that allows instructions to be inserted into binary. These instructions are inserted at a free place in the binary.
    Then, At the address given, an instruction is inserted that jumps to this block (also in the block are the instructions this overwrites).
    At the end of the block, it jumps back to right after the initial jump. The initial jump must be able to be inserted within the basic block
    of the given address.
    """

    def __init__(
        self,
        addr_or_name: int | str,
        instr: str,
        force_insert=False,
        detour_pos=-1,
        symbols: dict[str, int] | None = None,
        is_thumb=False,
        language: str = "ASM",
        c_config: CConfig | None = None,
        **kwargs,
    ) -> None:
        """
        Constructor.

        :param addr_or_name: If an integer, the new instructions are placed in a free spot in the binary and the jump to them is inserted at that memory address.
                             If a string, the new instructions are placed in a free spot in the binary and added as a symbol (with this as its name).
        :param instr: Instructions to insert. You can use "SAVE_CONTEXT" and "RESTORE_CONTEXT" wherever you want to save and restore program context. If you want to use any symbols from the program or from previous patches, you must surround them with curly braces.
        :param force_insert: If Patcherex should ignore whether instructions can be moved when inserting, defaults to False
        :param detour_pos: If given a name, specifies the file address to place the new instructions, defaults to -1
        :param symbols: Symbols to include when assembling, in format {symbol name: memory address}, defaults to None
        :param is_thumb: Whether the instructions given are thumb, defaults to False
        :param language: The language of the patch, can be either "ASM" or "C"
        :param c_config: Configuration options for when language == "C"
        :param **kwargs: Extra options. Can have a boolean "save_context" for whether context should be saved.
        """
        self.addr = None
        self.name = None
        if isinstance(addr_or_name, int):
            self.addr = addr_or_name
        elif isinstance(addr_or_name, str):
            self.name = addr_or_name
        self.instr = instr
        self.force_insert = force_insert
        self.detour_pos = detour_pos
        self.symbols = symbols if symbols else {}
        self.is_thumb = is_thumb
        self.language = language.upper()
        self.c_config = self.CConfig() if c_config is None else c_config
        self.save_context = (
            kwargs["save_context"] if "save_context" in kwargs else False
        )
        self.compile_opts = kwargs["compile_opts"] if "compile_opts" in kwargs else {}

    def apply(self, p):
        """
        Applies the patch to the binary, intended to be called by a Patcherex instance.

        :param p: Patcherex instance.
        :type p: Patcherex
        """
        if self.language == "ASM":
            self._apply_asm(p)
        elif self.language == "C":
            self._apply_c(p)

    def _apply_c(self, p) -> None:
        if self.addr is None:
            raise ValueError("An address must be provided for a C instruction patch")

        c_forward_header = self.c_config.c_forward_header
        c_scratch_regs = (
            frozenset()
            if self.c_config.scratch_regs is None
            else frozenset(self.c_config.scratch_regs)
        )
        c_regs_sort = (
            [] if self.c_config.regs_sort is None else list(self.c_config.regs_sort)
        )

        # TODO: make it os agnostic?
        calling_convention = (
            p.archinfo.cc["LinuxPreserveNone"]
            if p.compiler.preserve_none
            else p.archinfo.cc["Linux"]
        )
        subregister_table = p.archinfo.subregisters
        subregister_float_table = p.archinfo.subregisters_float

        # Figure out if there are any extra registers that we need to expose to the user
        # that aren't part of the calling convention. For x64 preserve_none, this will be
        # registers r10 and rbx.
        extra_saved = set(p.archinfo.regs)
        # Note that we cannot control callee saved registers. If we attempt to define
        # some registers via 'register uint64_t rbx asm("rbx");', the compiler will insert
        # push and pop instructions to save these registers.
        extra_saved = (
            extra_saved
            - set(calling_convention)
            - set(p.archinfo.callee_saved["Linux"])
        )
        extra_saved_in = list(extra_saved)
        # We don't want to necessarily output registers that have been marked as scratch
        # However we always want to make them available as input
        extra_saved_out = list(extra_saved - c_scratch_regs)

        def uint_converter(size: int):
            return f"uint{size}_t"

        extra_saved_in_converted = convert_to_subregisters(
            extra_saved_in, subregister_table, c_regs_sort, uint_converter
        )
        extra_saved_out_converted = convert_to_subregisters(
            extra_saved_out, subregister_table, c_regs_sort, uint_converter
        )

        calling_convention_float: list[str] = p.archinfo.cc_float["Linux"]
        extra_saved_float = set(p.archinfo.regs_float)
        extra_saved_float = (
            extra_saved_float
            - set(calling_convention_float)
            - set(p.archinfo.callee_saved_float["Linux"])
        )
        extra_saved_float_in = list(extra_saved_float)
        extra_saved_float_out = list(extra_saved_float - c_scratch_regs)

        def float_converter(size: int):
            if size in p.archinfo.float_types:
                return p.archinfo.float_types[size]
            else:
                raise ValueError(
                    f"Unable to determine type of float with size of {size} bits"
                )

        extra_saved_float_in_converted = convert_to_subregisters(
            extra_saved_float_in, subregister_float_table, c_regs_sort, float_converter
        )
        extra_saved_float_out_converted = convert_to_subregisters(
            extra_saved_float_out, subregister_float_table, c_regs_sort, float_converter
        )

        attribute = "__attribute__((preserve_none))" if p.compiler.preserve_none else ""

        int_args = convert_to_subregisters(
            calling_convention, subregister_table, c_regs_sort, uint_converter
        )
        int_args_str = [f"{itype} {name}" for (name, itype) in int_args]
        float_args: list[tuple[str, str]] = convert_to_subregisters(
            calling_convention_float,
            subregister_float_table,
            c_regs_sort,
            float_converter,
        )
        float_args_str = [f"{ftype} {name}" for (name, ftype) in float_args]
        args_str = ", ".join(int_args_str + float_args_str)

        callback_forward_decl = f"extern void {attribute} _CALLBACK({args_str});"

        # Stupid macro tricks to make coding the patch a little bit nicer. This allows the user
        # to write 'return;' instead of having to understand how to call the callback
        # reg_name if reg_name in self.c_out_regs else '_dummy'
        return_macro_lines = [
            "#define return do {",
        ]

        for reg, _ in extra_saved_out_converted:
            # Make sure the variables are live just before the return statement
            return_macro_lines.append(f'    asm ("" : : "r"({reg}) :);')
        for reg, _ in extra_saved_float_out_converted:
            # Make sure the variables are live just before the return statement
            return_macro_lines.append(f'    asm ("" : : "r"({reg}) :);')

        callback_args = [
            "_dummy" if reg_name in c_scratch_regs else reg_name
            for (reg_name, bits) in int_args
        ]
        callback_args += [
            "_dummyFloat" if reg_name in c_scratch_regs else reg_name
            for (reg_name, ftype) in float_args
        ]
        return_macro_lines += [
            "    __attribute__((musttail)) return _CALLBACK({});".format(
                ", ".join(callback_args)
            ),
            "} while(0)",
        ]

        return_macro = "\\\n".join(return_macro_lines)

        lines = [
            "#include <stdint.h>",
            "",
            callback_forward_decl,
            "",
            c_forward_header,
            "",
            return_macro,
            "",
            f"void {attribute} _MICROPATCH({args_str}) {{",
            f"    uint{p.archinfo.bits}_t _dummy;",
            "    float _dummyFloat;",
        ]
        for reg, reg_type in extra_saved_in_converted:
            # Force the variables to live in a specific register using the register C extension
            lines.append(f'    register {reg_type} {reg} asm("{reg}");')
        for reg, ftype in extra_saved_float_in_converted:
            lines.append(f'    register {ftype} {reg} asm("{reg}");')
        for reg, _ in extra_saved_in_converted:
            # Trick the C compiler into thinking that the variables we just defined are actually live
            lines.append(f'    asm ("" : "=r"({reg}) : : );')
        for reg, _ in extra_saved_float_in_converted:
            lines.append(f'    asm ("" : "=r"({reg}) : : );')
        lines += [
            self.instr,
            # Make sure we actually do the callback in case the user forgets to put in a return
            "    return;",
            "}",
            "#undef return",
        ]
        code = "\n".join(lines)
        logger.info("InsertInstructionPatch generated C code:\n" + code)
        p.utils.insert_trampoline_code(
            self.addr,
            code,
            force_insert=self.force_insert,
            detour_pos=self.detour_pos,
            symbols=self.symbols,
            language="C",
            asm_header=self.c_config.asm_header,
            asm_footer=self.c_config.asm_footer,
        )

    def _apply_asm(self, p) -> None:
        if self.addr:
            if "SAVE_CONTEXT" in self.instr:
                self.instr = self.instr.replace(
                    "SAVE_CONTEXT", f"\n{p.archinfo.save_context_asm}\n"
                )
            if "RESTORE_CONTEXT" in self.instr:
                self.instr = self.instr.replace(
                    "RESTORE_CONTEXT", f"\n{p.archinfo.restore_context_asm}\n"
                )
            if self.save_context:
                self.instr = f"{p.archinfo.save_context_asm}\n{self.instr}\n{p.archinfo.restore_context_asm}"
            p.utils.insert_trampoline_code(
                self.addr,
                self.instr,
                force_insert=self.force_insert,
                detour_pos=self.detour_pos,
                symbols=self.symbols,
            )
        elif self.name:
            assembled_size = len(
                p.assembler.assemble(
                    self.instr, symbols=self.symbols, is_thumb=self.is_thumb
                )
            )
            if self.detour_pos == -1:
                block = p.allocation_manager.allocate(
                    assembled_size, align=p.archinfo.alignment, flag=MemoryFlag.RX
                )
                p.symbols[self.name] = block.mem_addr
                p.binfmt_tool.update_binary_content(
                    block.file_addr,
                    p.assembler.assemble(
                        self.instr,
                        block.mem_addr,
                        symbols=self.symbols,
                        is_thumb=self.is_thumb,
                    ),
                )
            else:
                p.symbols[self.name] = self.detour_pos
                p.binfmt_tool.update_binary_content(
                    self.detour_pos,
                    p.assembler.assemble(
                        self.instr,
                        self.detour_pos,
                        symbols=self.symbols,
                        is_thumb=self.is_thumb,
                    ),
                )


class RemoveInstructionPatch(Patch):
    """
    Patch that removes instructions in the binary. Currently only takes in a number of bytes and an starting address.
    The number of bytes must be divisible by the nop size of the architecture, otherwise it will fail.
    """

    def __init__(
        self,
        addr: int,
        num_instr: int | None = None,
        num_bytes: int | None = None,
    ) -> None:
        """
        Constructor.

        :param addr: Memory address to remove instructions at.
        :param num_instr: Number of instructions to remove, currently not used, defaults to None
        :param num_bytes: Number of bytes to remove, must be divisible by nop size, defaults to None
        """
        self.addr = addr
        self.num_instr = num_instr
        self.num_bytes = num_bytes
        if self.num_instr is None and self.num_bytes is None:
            self.num_instr = 1

    def apply(self, p: Patcherex) -> None:
        """
        Applies the patch to the binary, intended to be called by a Patcherex instance.

        :param p: Patcherex instance.
        """
        if self.num_bytes is None:
            raise NotImplementedError()
        if self.num_bytes and self.num_bytes % p.archinfo.nop_size != 0:
            raise Exception(
                f"Cannot remove {self.num_bytes} bytes, must be a multiple of {p.archinfo.nop_size}"
            )
        num_nops = self.num_bytes // p.archinfo.nop_size
        offset = p.binary_analyzer.mem_addr_to_file_offset(self.addr)
        p.binfmt_tool.update_binary_content(offset, p.archinfo.nop_bytes * num_nops)



from __future__ import annotations

import json
import logging
import os
import subprocess
import tempfile

import cle
from elftools.elf.elffile import ELFFile

from ..assets.assets import Assets
from .clang import Clang

logger = logging.getLogger(__name__)


class LLVMRecomp(Clang):
    def __init__(
        self, p, clang_version=15, compiler_flags: list[str] | None = None
    ) -> None:
        super().__init__(p, clang_version, compiler_flags)
        self._clang_version = clang_version
        self._assets_path = Assets("llvm_recomp").path

    def compile(
        self,
        code: str,
        base=0,
        symbols: dict[str, int] | None = None,
        extra_compiler_flags: list[str] | None = None,
        is_thumb=False,
        **kwargs,
    ) -> bytes:
        if symbols is None:
            symbols = {}
        if extra_compiler_flags is None:
            extra_compiler_flags = []
        with tempfile.TemporaryDirectory() as td:
            # source file
            with open(os.path.join(td, "code.c"), "w") as f:
                f.write(code)

            librecomp_path = os.path.join(self._assets_path, "libRecompiler.so")

            # c -> ll
            try:
                args = (
                    [self._compiler]
                    + self._compiler_flags
                    + extra_compiler_flags
                    + [
                        "-Wno-incompatible-library-redeclaration",
                        "-S",
                        "-w",
                        "-emit-llvm",
                        "-g",
                        "-o",
                        os.path.join(td, "code.ll"),
                        os.path.join(td, "code.c"),
                        "-I/usr/lib/clang/15/include",
                    ]
                )
                subprocess.run(args, check=True, capture_output=True)
            except subprocess.CalledProcessError as e:
                logger.error(e.stderr.decode("utf-8"))
                raise e

            # ll --force-dso-local --> ll
            if "dso_local_fix" in kwargs and kwargs["dso_local_fix"]:
                try:
                    args = [
                        f"opt-{self._clang_version}",
                        f"-load-pass-plugin={librecomp_path}",
                        "-passes=force-dso-local",
                        "-S",
                        os.path.join(td, "code.ll"),
                        "-o",
                        os.path.join(td, "code.ll"),
                    ]
                    subprocess.run(args, check=True, capture_output=True)
                except subprocess.CalledProcessError as e:
                    logger.error(e.stderr.decode("utf-8"))
                    raise e

            # ll -> o
            if "stacklayout" in kwargs and kwargs["stacklayout"]:
                with open(os.path.join(td, "stacklayout.json"), "w") as f:
                    json.dump(kwargs["stacklayout"], f)
                try:
                    args = [
                        f"llc-{self._clang_version}",
                        "-stop-before=prologepilog",
                        os.path.join(td, "code.ll"),
                        "-o",
                        os.path.join(td, "code.mir"),
                        "-relocation-model=pic",
                    ]
                    subprocess.run(args, check=True, capture_output=True)
                except subprocess.CalledProcessError as e:
                    logger.error(e.stderr.decode("utf-8"))
                    raise e
                try:
                    args = [
                        f"llc-{self._clang_version}",
                        "-load",
                        librecomp_path,
                        "-run-pass=updated-prologepilog",
                        f"-stkloc={os.path.join(td, 'stacklayout.json')}",
                        "-o",
                        os.path.join(td, "code.2.mir"),
                        os.path.join(td, "code.mir"),
                        "-relocation-model=pic",
                    ]
                    subprocess.run(args, check=True, capture_output=True)
                except subprocess.CalledProcessError as e:
                    logger.error(e.stderr.decode("utf-8"))
                    raise e
                try:
                    args = [
                        f"llc-{self._clang_version}",
                        "-start-after=prologepilog",
                        "-o",
                        os.path.join(td, "obj.o"),
                        os.path.join(td, "code.2.mir"),
                        "-relocation-model=pic",
                        "--filetype=obj",
                    ]
                    subprocess.run(args, check=True, capture_output=True)
                except subprocess.CalledProcessError as e:
                    logger.error(e.stderr.decode("utf-8"))
                    raise e
            else:
                try:
                    args = [
                        f"llc-{self._clang_version}",
                        "-o",
                        os.path.join(td, "obj.o"),
                        os.path.join(td, "code.ll"),
                        "-relocation-model=pic",
                        "--filetype=obj",
                    ]
                    subprocess.run(args, check=True, capture_output=True)
                except subprocess.CalledProcessError as e:
                    logger.error(e.stderr.decode("utf-8"))
                    raise e

            # linker script
            _symbols = {}
            _symbols.update(self.p.symbols)
            _symbols.update(self.p.binary_analyzer.get_all_symbols())
            _symbols.update(symbols)

            with open(os.path.join(td, "obj.o"), "rb") as f:
                elf = ELFFile(f)
                linker_script_rodata_sections = " ".join(
                    [
                        f". = ALIGN({section['sh_addralign']}); *({section.name})"
                        for section in elf.iter_sections()
                        if section.name.startswith(".rodata")
                    ]
                )

                # automatically add symbols like off_deadbeef, dword_deadbeef, etc.
                for sym in elf.get_section_by_name(".symtab").iter_symbols():
                    if (
                        sym.entry.st_shndx == "SHN_UNDEF"
                        and sym.name
                        and "_" in sym.name
                    ):
                        try:
                            _, addr = sym.name.split("_", 1)
                            addr = int(addr, 16)
                            if sym.name not in _symbols:
                                _symbols[sym.name] = addr
                        except ValueError:
                            pass
            linker_script_symbols = "".join(
                f"{name} = {hex(addr)};" for name, addr in _symbols.items()
            )

            linker_script = f"SECTIONS {{ .patcherex2 : SUBALIGN(0) {{ . = {hex(base)}; *(.text) {linker_script_rodata_sections} {linker_script_symbols} }} }}"
            with open(os.path.join(td, "linker.ld"), "w") as f:
                f.write(linker_script)

            # link object file
            try:
                args = [self._linker] + [
                    "-relocatable",
                    os.path.join(td, "obj.o"),
                    "-T",
                    os.path.join(td, "linker.ld"),
                    "-o",
                    os.path.join(td, "obj_linked.o"),
                ]
                subprocess.run(args, check=True, capture_output=True)
            except subprocess.CalledProcessError as e:
                logger.error(e.stderr.decode("utf-8"))
                raise e

            # extract compiled code
            ld = cle.Loader(
                os.path.join(td, "obj_linked.o"), main_opts={"base_addr": 0x0}
            )

            patcherex2_section = next(
                (s for s in ld.main_object.sections if s.name == ".patcherex2"), None
            )
            compiled_start = ld.all_objects[0].entry + base

            compiled = ld.memory.load(
                compiled_start,
                patcherex2_section.memsize - compiled_start,
            )
        return compiled



import logging

logger = logging.getLogger(__name__)


class Assembler:
    def __init__(self, p) -> None:
        self.p = p

    def resolve_symbols(self, code: str, symbols=None):
        if symbols is None:
            symbols = {}
        _symbols = {}
        _symbols.update(self.p.symbols)
        _symbols.update(self.p.binary_analyzer.get_all_symbols())
        _symbols.update(symbols)

        for symbol, addr in _symbols.items():
            code = code.replace(f"{{{symbol}}}", hex(addr))
        return code

    def _assemble(self, code: str, base=0, **kwargs) -> None:
        raise NotImplementedError()

    def _pre_assemble_hook(self, code: str, base=0) -> None:
        return code

    def assemble(self, code: str, base=0, symbols=None, **kwargs) -> None:
        if code == "":
            return b""
        if symbols is None:
            symbols = {}
        logger.debug(f"Assembling `{code}` at {hex(base)}")
        code = self.resolve_symbols(code, symbols=symbols)
        code = self._pre_assemble_hook(code, base=base)

        return self._assemble(code, base=base, **kwargs)



import logging
import os

from ..assets.assets import Assets
from .compiler import Compiler

logger = logging.getLogger(__name__)


class PpcVle(Compiler):
    def __init__(self, p) -> None:
        super().__init__(p)
        self.assets_path = Assets("ppc_vle").path
        self._compiler = os.path.join(self.assets_path, "powerpc-eabivle-gcc")
        self._linker = os.path.join(self.assets_path, "powerpc-eabivle-ld")
        self._compiler_flags = ["-mno-vle"]



from __future__ import annotations

from headless_ida import HeadlessIdaRemote

from .binary_analyzer import BinaryAnalyzer


class Ida(BinaryAnalyzer):
    _DEFAULT_LOAD_BASE = 0x0

    def __init__(self, binary_path: str, **kwargs) -> None:
        self.binary_path = binary_path
        self.kwargs = kwargs
        self.ida_server_host = (
            kwargs["ida_server_host"] if "ida_server_host" in kwargs else "localhost"
        )
        self.ida_server_port = (
            kwargs["ida_server_port"] if "ida_server_port" in kwargs else 1337
        )
        self._headlessida = HeadlessIdaRemote(
            self.ida_server_host, self.ida_server_port, self.binary_path
        )
        ida_libs = [
            "idc",
            "idautils",
            "idaapi",
            "ida_funcs",
            "ida_xref",
            "ida_nalt",
            "ida_auto",
            "ida_hexrays",
            "ida_name",
            "ida_expr",
            "ida_struct",
            "ida_typeinf",
            "ida_loader",
            "ida_lines",
            "ida_segment",
            "ida_gdl",
        ]
        for lib in ida_libs:
            setattr(self, lib, self._headlessida.import_module(lib))

    def mem_addr_to_file_offset(self, addr: int) -> int:
        return self.ida_loader.get_fileregion_offset(addr)

    def get_basic_block(self, addr: int) -> dict[str, int | list[int]]:
        func = self.ida_funcs.get_func(addr)
        instr_addrs = list(func.code_items())
        assert addr in instr_addrs, "Invalid address"
        flowchart = self.ida_gdl.FlowChart(f=func, flags=self.ida_gdl.FC_PREDS)

        for block in flowchart:
            if block.start_ea <= addr < block.end_ea:
                return {
                    "start": block.start_ea,
                    "end": block.end_ea,
                    "size": block.end_ea - block.start_ea,
                    "instruction_addrs": [
                        ea for ea in instr_addrs if block.start_ea <= ea < block.end_ea
                    ],
                }

    def get_instr_bytes_at(self, addr: int):
        return self.p.factory.block(addr, num_inst=1).bytes



from ..components.allocation_managers.allocation_manager import AllocationManager
from ..components.archinfo.mips64 import Mips64Info
from ..components.assemblers.keystone import Keystone, keystone
from ..components.binary_analyzers.angr import Angr
from ..components.binary_analyzers.ghidra import Ghidra
from ..components.binfmt_tools.elf import ELF
from ..components.compilers.clang import Clang
from ..components.disassemblers.capstone import Capstone, capstone
from ..components.utils.utils import Utils
from .target import Target


class ElfMips64Linux(Target):
    @staticmethod
    def detect_target(binary_path):
        with open(binary_path, "rb") as f:
            magic = f.read(0x14)
            if magic.startswith(b"\x7fELF\x02") and magic.startswith(
                b"\x00\x08", 0x12
            ):  # EM_MIPS
                return True
        return False

    def get_assembler(self, assembler):
        assembler = assembler or "keystone"
        if assembler == "keystone":
            return Keystone(
                self.p,
                keystone.KS_ARCH_MIPS,
                keystone.KS_MODE_BIG_ENDIAN + keystone.KS_MODE_64,
            )
        raise NotImplementedError()

    def get_allocation_manager(self, allocation_manager):
        allocation_manager = allocation_manager or "default"
        if allocation_manager == "default":
            return AllocationManager(self.p)
        raise NotImplementedError()

    def get_compiler(self, compiler):
        compiler = compiler or "clang"
        if compiler == "clang":
            return Clang(self.p, compiler_flags=["--target=mips64-linux-gnuabi64"])
        raise NotImplementedError()

    def get_disassembler(self, disassembler):
        disassembler = disassembler or "capstone"
        if disassembler == "capstone":
            return Capstone(
                capstone.CS_ARCH_MIPS,
                capstone.CS_MODE_BIG_ENDIAN + capstone.CS_MODE_MIPS64,
            )
        raise NotImplementedError()

    def get_binfmt_tool(self, binfmt_tool):
        binfmt_tool = binfmt_tool or "pyelftools"
        if binfmt_tool == "pyelftools":
            return ELF(self.p, self.binary_path)
        raise NotImplementedError()

    def get_binary_analyzer(self, binary_analyzer, **kwargs):
        binary_analyzer = binary_analyzer or "angr"
        if binary_analyzer == "angr":
            return Angr(self.binary_path, **kwargs)
        if binary_analyzer == "ghidra":
            return Ghidra(self.binary_path, **kwargs)
        raise NotImplementedError()

    def get_utils(self, utils):
        utils = utils or "default"
        if utils == "default":
            return Utils(self.p, self.binary_path)
        raise NotImplementedError()

    def get_archinfo(self, archinfo):
        archinfo = archinfo or "default"
        if archinfo == "default":
            return Mips64Info()
        raise NotImplementedError()



# ruff: noqa: F403, F405
from __future__ import annotations

import logging

from .components.binary_analyzers.ghidra import Ghidra
from .patches import *
from .patches import __all__ as all_patches
from .targets import Target

logging.Logger.manager.loggerDict["patcherex"] = logging.Logger.manager.loggerDict[
    "patcherex2"
]
logger = logging.getLogger(__name__)


class Patcherex:
    """
    The main class of the library. This is how you are intended to interact with patches.
    """

    def __init__(
        self,
        binary_path: str,
        target_cls: type[Target] | None = None,
        target_opts: dict[str, str] | None = None,
        components_opts: dict[str, dict[str, str]] | None = None,
    ) -> None:
        """
        Constructor.

        :param binary_path: The path of the binary to patch.
        :param target_cls: Specified architecture class to use, otherwise it is automatically detected, defaults to None
        :param target_opts: Options to specify components for the target, defaults to None
        :param components_opts: Options for configuring each component for the target, defaults to None
        """
        if target_opts is None:
            target_opts = {}
        if components_opts is None:
            components_opts = {}
        self.binary_path = binary_path
        if target_cls is None:
            self.target = Target.detect_target(self, binary_path)
        else:
            self.target = target_cls(self, binary_path)

        self.symbols = {}
        self.sypy_info = {"patcherex_added_functions": []}
        self.patches = []

        # Initialize components
        components = [
            "assembler",
            "disassembler",
            "compiler",
            "binary_analyzer",
            "allocation_manager",
            "binfmt_tool",
            "utils",
            "archinfo",
        ]
        for component in components:
            setattr(
                self,
                component,
                self.target.get_component(
                    component,
                    target_opts.get(component),
                    components_opts.get(component),
                ),
            )

        # Chosen patch order, making sure all are accounted for
        self.patch_order = (
            ModifyRawBytesPatch,
            RemoveDataPatch,
            InsertDataPatch,
            ModifyDataPatch,
            RemoveLabelPatch,
            ModifyLabelPatch,
            InsertLabelPatch,
            RemoveInstructionPatch,
            InsertInstructionPatch,
            ModifyInstructionPatch,
            RemoveFunctionPatch,
            InsertFunctionPatch,
            ModifyFunctionPatch,
        )
        assert len(self.patch_order) == len(all_patches)

    def shutdown(self):
        """
        Shuts down any resources used by Patcherex2.
        This needs to be called when using Ghidra as the binary analyzer when done patching.
        """
        if isinstance(self.binary_analyzer, Ghidra):
            self.binary_analyzer.shutdown()

    def apply_patches(self) -> None:
        """
        Applies all added patches to the binary. Call this when you have added all the patches you want.
        """
        # TODO: sort patches properly
        # self.patches.sort(key=lambda x: self.patch_order.index(type(x)))
        self.patches.sort(
            key=lambda x: not isinstance(
                x, (ModifyDataPatch, InsertDataPatch, RemoveDataPatch)
            )
        )
        logger.debug(f"Applying patches: {self.patches}")
        for patch in self.patches:
            patch.apply(self)
        self.binfmt_tool.finalize()

    def save_binary(self, filename: str = None) -> None:
        """
        Save the patched binary to a file.

        :param filename: Name of file to save to, defaults to '<filename>.patched'
        """
        self.binfmt_tool.save_binary(filename)



import logging

from ..components.allocation_managers.allocation_manager import (
    AllocationManager,
    MappedBlock,
    MemoryFlag,
)
from ..components.archinfo.sparc import SparcInfo
from ..components.assemblers.bcc import Bcc as BccAssembler
from ..components.assemblers.keystone_sparc import KeystoneSparc, keystone
from ..components.binary_analyzers.angr import Angr
from ..components.binary_analyzers.ghidra import Ghidra
from ..components.binfmt_tools.elf import ELF
from ..components.compilers.bcc import Bcc as BccCompiler
from ..components.disassemblers.capstone import Capstone, capstone
from ..components.utils.utils import Utils
from .target import Target

logger = logging.getLogger(__name__)


class CustomElf(ELF):
    def _init_memory_analysis(self):
        # remove all non-RWX segments
        self._segments = [s for s in self._segments if s["p_flags"] & 0b111 == 0b111]
        block = MappedBlock(
            self._segments[0]["p_offset"],
            self._segments[0]["p_vaddr"],
            self._segments[0]["p_memsz"],
            is_free=False,
            flag=MemoryFlag.RWX,
        )
        self.p.allocation_manager.add_block(block)


class ElfLeon3Bare(Target):
    @staticmethod
    def detect_target(binary_path):
        return False

    def get_assembler(self, assembler):
        assembler = assembler or "keystone"
        if assembler == "keystone":
            return KeystoneSparc(
                self.p,
                keystone.KS_ARCH_SPARC,
                keystone.KS_MODE_SPARC32 + keystone.KS_MODE_BIG_ENDIAN,
            )
        elif assembler == "bcc":
            return BccAssembler(self.p)
        raise NotImplementedError()

    def get_allocation_manager(self, allocation_manager):
        allocation_manager = allocation_manager or "default"
        if allocation_manager == "default":
            return AllocationManager(self.p)
        raise NotImplementedError()

    def get_compiler(self, compiler):
        compiler = compiler or "bcc"
        if compiler == "bcc":
            return BccCompiler(self.p)
        raise NotImplementedError()

    def get_disassembler(self, disassembler):
        disassembler = disassembler or "capstone"
        if disassembler == "capstone":
            return Capstone(capstone.CS_ARCH_SPARC, capstone.CS_MODE_BIG_ENDIAN)
        raise NotImplementedError()

    def get_binfmt_tool(self, binfmt_tool):
        binfmt_tool = binfmt_tool or "custom"
        if binfmt_tool == "custom":
            return CustomElf(self.p, self.binary_path)
        raise NotImplementedError()

    def get_binary_analyzer(self, binary_analyzer, **kwargs):
        binary_analyzer = binary_analyzer or "angr"
        if binary_analyzer == "angr":
            return Angr(self.binary_path, **kwargs)
        if binary_analyzer == "ghidra":
            return Ghidra(self.binary_path, **kwargs)
        raise NotImplementedError()

    def get_utils(self, utils):
        utils = utils or "default"
        if utils == "default":
            return Utils(self.p, self.binary_path)
        raise NotImplementedError()

    def get_archinfo(self, archinfo):
        archinfo = archinfo or "default"
        if archinfo == "default":
            return SparcInfo()
        raise NotImplementedError()



import logging

import keystone

from .assembler import Assembler

logger = logging.getLogger(__name__)


class KeystoneArm(Assembler):
    def __init__(self, p) -> None:
        super().__init__(p)
        self.ks_arm = keystone.Ks(
            keystone.KS_ARCH_ARM, keystone.KS_MODE_ARM + keystone.KS_MODE_LITTLE_ENDIAN
        )
        self.ks_thumb = keystone.Ks(
            keystone.KS_ARCH_ARM,
            keystone.KS_MODE_THUMB + keystone.KS_MODE_LITTLE_ENDIAN,
        )

    def _assemble(self, code: str, base=0, is_thumb=False) -> bytes:
        try:
            ks = self.ks_thumb if is_thumb else self.ks_arm
            binary, _ = ks.asm(code, base)
            logger.debug(f"Assembled bytes: {bytes(binary).hex()}")
            return bytes(binary)
        except Exception as e:
            raise Exception(
                f'Failed to assemble: """\n{code}\n"""\nat base: {hex(base)}'
            ) from e



import logging

import archinfo

from ..components.allocation_managers.allocation_manager import AllocationManager
from ..components.archinfo.ppc_vle import PpcVleInfo
from ..components.assemblers.ppc_vle import PpcVle as PpcVleAssembler
from ..components.binary_analyzers.angr import Angr
from ..components.binfmt_tools.ihex import IHex
from ..components.compilers.ppc_vle import PpcVle as PpcVleCompiler
from ..components.disassemblers.ppc_vle import PpcVle as PpcVleDisassembler
from ..components.utils.utils import Utils
from .target import Target

logger = logging.getLogger(__name__)


class IHexPPCBare(Target):
    @staticmethod
    def detect_target(binary_path):
        return False

    def get_assembler(self, assembler):
        assembler = assembler or "default"
        if assembler == "default":
            return PpcVleAssembler(self.p)
        raise NotImplementedError()

    def get_allocation_manager(self, allocation_manager):
        allocation_manager = allocation_manager or "default"
        if allocation_manager == "default":
            return AllocationManager(self.p)
        raise NotImplementedError()

    def get_compiler(self, compiler):
        compiler = compiler or "default"
        if compiler == "default":
            return PpcVleCompiler(self.p)
        raise NotImplementedError()

    def get_disassembler(self, disassembler):
        disassembler = disassembler or "default"
        if disassembler == "default":
            return PpcVleDisassembler(self.p)
        raise NotImplementedError()

    def get_binfmt_tool(self, binfmt_tool):
        binfmt_tool = binfmt_tool or "default"
        if binfmt_tool == "default":
            return IHex(self.p, self.binary_path)
        raise NotImplementedError()

    def get_binary_analyzer(self, binary_analyzer):
        binary_analyzer = binary_analyzer or "angr"
        if binary_analyzer == "angr":
            return Angr(
                self.binary_path,
                angr_kwargs={
                    "arch": archinfo.ArchPcode("PowerPC:BE:32:MPC8270"),
                    "auto_load_libs": False,
                    "load_debug_info": True,
                },
                angr_cfg_kwargs={
                    "normalize": True,
                    "data_references": True,
                    "force_smart_scan": False,
                    "force_complete_scan": False,
                },
            )
        raise NotImplementedError()

    def get_utils(self, utils):
        utils = utils or "default"
        if utils == "default":
            return Utils(self.p, self.binary_path)
        raise NotImplementedError()

    def get_archinfo(self, archinfo):
        archinfo = archinfo or "default"
        if archinfo == "default":
            return PpcVleInfo()
        raise NotImplementedError()



class Patch:
    """
    Base class for patches. Not instantiated direactly.
    """

    def __init__(self, parent=None) -> None:
        self.parent = parent

    def apply(self, p):
        raise NotImplementedError()



from .patch import Patch


class InsertLabelPatch(Patch):
    def __init__(self, addr: int) -> None:
        self.addr = addr
        raise NotImplementedError()


class ModifyLabelPatch(Patch):
    def __init__(self, addr: int) -> None:
        self.addr = addr
        raise NotImplementedError()


class RemoveLabelPatch(Patch):
    def __init__(self, addr: int) -> None:
        self.addr = addr
        raise NotImplementedError()



class ArmInfo:
    nop_bytes = b"\x00\xf0\x20\xe3"  # TODO: thumb
    nop_size = 4
    jmp_asm = "b {dst}"
    jmp_size = 4
    alignment = 4
    bits = 32
    is_variable_length_isa = False
    instr_size = 4  # TODO: thumb 2
    call_asm = "bl {dst}"
    pc_reg_names = ["pc", "r15"]
    save_context_asm = """
        push {r0-r11}
    """
    restore_context_asm = """
        pop {r0-r11}
    """

    cc = {
        "Linux": ["r0", "r1", "r2", "r3"],
        # TODO: update LinuxPreserveNone once aarch64 support lands
        # in LLVM for preserve_none, currently defaults to Linux
        "LinuxPreserveNone": ["r0", "r1", "r2", "r3"],
    }
    callee_saved = {
        "Linux": ["r4", "r5", "r6", "r7", "r8", "r9", "r10", "r11", "r13", "r14", "r15"]
    }
    cc_float = {"Linux": ["d0", "d1", "d2", "d3", "d4", "d5", "d6", "d7"]}
    callee_saved_float = {
        "Linux": ["d8", "d9", "d10", "d11", "d12", "d13", "d14", "d15"]
    }

    float_types = {32: "float", 64: "double"}

    @property
    def regs(self):
        return list(self.subregisters.keys())

    @property
    def regs_float(self):
        return list(self.subregisters_float.keys())

    subregisters = {f"r{i}": {32: [f"r{i}"]} for i in range(0, 16)}

    subregisters_float = {}
    for i in range(0, 16):
        subregisters_float[f"d{i}"] = {64: [f"d{i}"], 32: [f"s{2 * i}"]}
    for i in range(16, 32):
        subregisters_float[f"d{i}"] = {64: [f"d{i}"]}



import logging
import os

from ..assets.assets import Assets
from .compiler import Compiler

logger = logging.getLogger(__name__)


class Bcc(Compiler):
    def __init__(self, p) -> None:
        super().__init__(p)
        self.assets_path = Assets("bcc").path
        self._compiler = os.path.join(self.assets_path, "sparc-gaisler-elf-gcc")
        self._linker = os.path.join(self.assets_path, "sparc-gaisler-elf-ld")
        self._compiler_flags = ["-qbsp=gr712rc", "-mcpu=leon3", "-mfix-gr712rc"]



from __future__ import annotations

import logging
import tempfile

from .binary_analyzer import BinaryAnalyzer

logger = logging.getLogger(__name__)


class Ghidra(BinaryAnalyzer):
    def __init__(self, binary_path: str, **kwargs):
        import pyhidra

        self.temp_proj_dir_ctx = tempfile.TemporaryDirectory()
        self.temp_proj_dir = self.temp_proj_dir_ctx.__enter__()

        self.pyhidra_ctx = pyhidra.open_program(binary_path, self.temp_proj_dir)
        self.flatapi = self.pyhidra_ctx.__enter__()
        self.currentProgram = self.flatapi.getCurrentProgram()

        import ghidra

        self.ghidra = ghidra

        self.bbm = self.ghidra.program.model.block.BasicBlockModel(self.currentProgram)

    def shutdown(self):
        self.pyhidra_ctx.__exit__(None, None, None)
        self.temp_proj_dir_ctx.__exit__(None, None, None)

    def normalize_addr(self, addr):
        addr = addr.getOffset()
        if self.currentProgram.getRelocationTable().isRelocatable():
            addr -= self.currentProgram.getImageBase().getOffset()
        return addr

    def denormalize_addr(self, addr):
        if self.currentProgram.getRelocationTable().isRelocatable():
            addr += self.currentProgram.getImageBase().getOffset()
        return self.flatapi.toAddr(hex(addr))

    def mem_addr_to_file_offset(self, addr: int) -> int:
        addr = self.denormalize_addr(addr)
        try:
            return (
                self.currentProgram.getMemory()
                .getAddressSourceInfo(addr)
                .getFileOffset()
            )
        except Exception:
            raise Exception("Can't get file offset for addr") from None

    def get_basic_block(self, addr: int) -> dict[str, int | list[int]]:
        logger.info(f"getting basic block at 0x{addr} with ghidra")
        addr = self.denormalize_addr(addr)

        block = self.bbm.getFirstCodeBlockContaining(
            addr, self.ghidra.util.task.TaskMonitor.DUMMY
        )
        if block is None:
            raise Exception(f"Cannot find block containing address 0x{addr}")
        instrs = []
        ii = self.currentProgram.getListing().getInstructions(block, True)
        for i in ii:
            instrs.append(self.normalize_addr(i.getAddress()))
        return {
            "start": self.normalize_addr(block.getMinAddress()),
            "end": self.normalize_addr(block.getMinAddress()) + block.getNumAddresses(),
            "size": block.getNumAddresses(),
            "instruction_addrs": instrs,
        }

    def get_instr_bytes_at(self, addr: int, num_instr=1):
        addr = self.denormalize_addr(addr)
        instr = self.currentProgram.getListing().getInstructionContaining(addr)
        if instr is None:
            return None
        b = instr.getBytes()
        for _i in range(1, num_instr):
            instr = instr.getNext()
            b = b"".join([b, instr.getBytes()])
        logger.info(
            f"got instr bytes of length {len(b)} for {num_instr} instrs at 0x{addr} with ghidra"
        )
        return b

    def get_unused_funcs(self) -> list[dict[str, int]]:
        logger.info("getting unused funcs with ghidra")
        fi = self.currentProgram.getListing().getFunctions(True)
        unused_funcs = []
        for f in fi:
            if not f.getSymbol().hasReferences():
                b = f.getBody()
                unused_funcs.append(
                    {
                        "addr": self.normalize_addr(b.getMinAddress()),
                        "size": b.getNumAddresses(),
                    }
                )
        return unused_funcs

    def get_all_symbols(self) -> dict[str, int]:
        logger.info("getting all symbols with ghidra")
        symbols = {}
        # si = self.currentProgram.getSymbolTable().getAllSymbols(True)
        # for s in si:
        #     if not s.isPrimary():
        #         continue
        #     symbols[s.getName()] = self.normalize_addr(
        #         s.getAddress().getOffset())
        fi = self.currentProgram.getListing().getFunctions(True)
        for f in fi:
            if f.getName() in symbols.keys():
                continue
            symbols[f.getName()] = self.normalize_addr(f.getEntryPoint())
            if self.is_thumb(symbols[f.getName()]):
                symbols[f.getName()] += 1
        return symbols

    def get_function(self, name_or_addr: int | str) -> dict[str, int] | None:
        if isinstance(name_or_addr, int):
            name_or_addr = self.denormalize_addr(name_or_addr)
            func = self.currentProgram.getListing().getFunctionContaining(name_or_addr)
            if func is None:
                return None
        elif isinstance(name_or_addr, str):
            funcs = self.currentProgram.getListing().getGlobalFunctions(name_or_addr)
            if len(funcs) == 0:
                return None
            func = funcs[0]
        else:
            raise Exception("Invalid type for argument")

        b = func.getBody()
        return {
            "addr": self.normalize_addr(b.getMinAddress()),
            "size": b.getNumAddresses(),
        }

    def is_thumb(self, addr: int) -> bool:
        addr = self.denormalize_addr(addr)
        r = self.currentProgram.getRegister("TMode")
        if r is None:
            return False
        v = self.currentProgram.getProgramContext().getRegisterValue(r, addr)
        t = v.unsignedValueIgnoreMask.intValue() == 1
        logger.info(f"address 0x{addr} {'is' if t else 'is not'} thumb from ghidra")
        return t



from ..components.binary_analyzers.angr import Angr
from ..components.binary_analyzers.ghidra import Ghidra
from ..components.compilers.llvm_recomp_arm import LLVMRecompArm
from .elf_arm_linux import ElfArmLinux


class ElfArmLinuxRecomp(ElfArmLinux):
    @staticmethod
    def detect_target(binary_path):
        return False

    def get_compiler(self, compiler):
        compiler = compiler or "llvm_recomp"
        if compiler == "llvm_recomp":
            return LLVMRecompArm(self.p)
        raise NotImplementedError()

    def get_binary_analyzer(self, binary_analyzer, **kwargs):
        binary_analyzer = binary_analyzer or "angr"
        if binary_analyzer == "angr":
            return Angr(self.binary_path, **kwargs)
        if binary_analyzer == "ghidra":
            return Ghidra(self.binary_path, **kwargs)
        raise NotImplementedError()



from __future__ import annotations

import logging
import re

from ..allocation_managers.allocation_manager import MemoryFlag

logger = logging.getLogger(__name__)


class Utils:
    def __init__(self, p, binary_path: str) -> None:
        self.p = p
        self.binary_path = binary_path

    def insert_trampoline_code(
        self,
        addr: int,
        instrs: str,
        force_insert=False,
        detour_pos=-1,
        symbols: dict[str, int] = None,
        language="ASM",
        asm_header="",
        asm_footer="",
    ) -> None:
        logger.debug(f"Inserting trampoline code at {hex(addr)}: {instrs}")
        symbols = symbols if symbols else {}
        assert force_insert or self.is_valid_insert_point(addr), (
            f"Cannot insert instruction at {hex(addr)}"
        )
        if not force_insert:
            moved_instrs = self.get_instrs_to_be_moved(addr)
            moved_instrs_len = len(
                self.p.assembler.assemble(
                    moved_instrs,
                    addr,  # TODO: we don't really need this addr, but better than 0x0 because 0x0 is too far away from the code
                    is_thumb=self.p.binary_analyzer.is_thumb(addr),
                )
            )
        else:
            moved_instrs = ""
            moved_instrs_len = len(
                self.p.assembler.assemble(
                    self.get_instrs_to_be_moved(addr, ignore_unmovable=True),
                    addr,  # TODO: we don't really need this addr, but better than 0x0 because 0x0 is too far away from the code
                    is_thumb=self.p.binary_analyzer.is_thumb(addr),
                )
            )
        trampoline_instrs_with_jump_back = (
            # Insert the asm footer here
            (asm_footer if language == "C" else instrs)
            + "\n"
            + moved_instrs
            + "\n"
            + self.p.archinfo.jmp_asm.format(dst=hex(addr + moved_instrs_len))
        )

        if language == "C":
            symbols_copy = dict(symbols)
            symbols_copy["_CALLBACK"] = 0
            asm_header_length = len(
                self.p.assembler.assemble(
                    asm_header,
                    addr,
                    # TODO: we don't really need this addr, but better than 0x0 because 0x0 is too far away from the code
                    is_thumb=self.p.binary_analyzer.is_thumb(addr),
                )
            )
            compiled_length = len(
                self.p.compiler.compile(
                    instrs,
                    symbols=symbols_copy,
                    is_thumb=self.p.binary_analyzer.is_thumb(addr),
                    # Some optimization needs to be used to squash down all this parameter trickery
                    # we are doing. -Os optimizes for space
                    extra_compiler_flags=["-Os"],
                )
            )
        else:
            asm_header_length = 0
            compiled_length = 0

        trampoline_size = (
            asm_header_length
            + compiled_length
            + len(
                self.p.assembler.assemble(
                    trampoline_instrs_with_jump_back,
                    addr,  # TODO: we don't really need this addr, but better than 0x0 because 0x0 is too far away from the code
                    symbols=symbols,
                    is_thumb=self.p.binary_analyzer.is_thumb(addr),
                )
            )
            + 4  # TODO: some time actual size is larger, but we need a better way to calculate it
        )

        if detour_pos == -1:
            trampoline_block = self.p.allocation_manager.allocate(
                trampoline_size, align=self.p.archinfo.alignment, flag=MemoryFlag.RX
            )
            logger.debug(f"Allocated trampoline block: {trampoline_block}")
            mem_addr = trampoline_block.mem_addr
            file_addr = trampoline_block.file_addr
        else:
            mem_addr = detour_pos
            for block in self.p.allocation_manager.new_mapped_blocks:
                if block.mem_addr == mem_addr:
                    file_addr = block.file_addr
                    break
            else:
                file_addr = self.p.binary_analyzer.mem_addr_to_file_offset(mem_addr)

        if language == "C":
            # Compile asm header
            compiled_asm_header = self.p.assembler.assemble(
                asm_header,
                mem_addr,
                symbols=symbols,
                is_thumb=self.p.binary_analyzer.is_thumb(addr),
            )

            # Compile C code
            symbols_copy = dict(symbols)
            symbols_copy["_CALLBACK"] = mem_addr + asm_header_length + compiled_length
            compiled_code = self.p.compiler.compile(
                instrs,
                base=mem_addr + asm_header_length,
                symbols=symbols_copy,
                is_thumb=self.p.binary_analyzer.is_thumb(addr),
                # Some optimization needs to be used to squash down all this parameter trickery
                # we are doing. -Os optimizes for space
                extra_compiler_flags=["-Os"],
            )
            if len(compiled_code) != compiled_length:
                raise RuntimeError(
                    f"When the compiled patch was compiled with base address {hex(mem_addr + asm_header_length)}, the length of the patch was {len(compiled_code)} bytes. This differs from the length of {compiled_length} bytes computed when the base address was 0."
                )
        else:
            compiled_asm_header = b""
            compiled_code = b""

        self.p.sypy_info["patcherex_added_functions"].append(hex(mem_addr))
        trampoline_bytes = (
            compiled_asm_header
            + compiled_code
            + self.p.assembler.assemble(
                trampoline_instrs_with_jump_back,
                mem_addr + asm_header_length + compiled_length,
                symbols=symbols,
                is_thumb=self.p.binary_analyzer.is_thumb(addr),
            )
        )
        self.p.binfmt_tool.update_binary_content(file_addr, trampoline_bytes)
        jmp_to_trampoline = self.p.assembler.assemble(
            self.p.archinfo.jmp_asm.format(dst=hex(mem_addr)),
            addr,
            is_thumb=self.p.binary_analyzer.is_thumb(addr),
        )
        self.p.binfmt_tool.update_binary_content(
            self.p.binary_analyzer.mem_addr_to_file_offset(addr), jmp_to_trampoline
        )

    def get_instrs_to_be_moved(self, addr: int, ignore_unmovable=False) -> str | None:
        basic_block = self.p.binary_analyzer.get_basic_block(addr)
        idx = basic_block["instruction_addrs"].index(addr)
        end = addr + self.p.archinfo.jmp_size
        instrs = b""

        for insn_addr in basic_block["instruction_addrs"][idx:] + [basic_block["end"]]:
            if end <= insn_addr:
                # we have enough space to insert a jump
                disasms = self.p.disassembler.disassemble(
                    instrs,
                    addr,
                    is_thumb=self.p.binary_analyzer.is_thumb(addr),
                )
                return "\n".join(
                    [self.p.disassembler.to_asm_string(d) for d in disasms]
                )
            if insn_addr == basic_block["end"]:
                # we reached the end of the basic block
                return None
            if not ignore_unmovable and not self.is_movable_instruction(insn_addr):
                logger.error(f"Instruction at {hex(insn_addr)} is not movable")
                # we cannot insert a jump here
                return None
            instrs += self.p.binary_analyzer.get_instr_bytes_at(insn_addr)
        return None

    def is_valid_insert_point(self, addr: int) -> bool:
        return self.get_instrs_to_be_moved(addr) is not None

    def is_movable_instruction(self, addr: int) -> bool:
        is_thumb = self.p.binary_analyzer.is_thumb(addr)
        insn_bytes = self.p.binary_analyzer.get_instr_bytes_at(addr)
        disassembled = self.p.disassembler.disassemble(
            insn_bytes, addr, is_thumb=is_thumb
        )[0]
        # if instruction use PC as a base register, it's not movable
        tokens = re.split(r"\s|,|\[|\]", disassembled["op_str"])
        tokens = list(filter(None, tokens))
        if list(set(self.p.archinfo.pc_reg_names) & set(tokens)):
            return False
        # TODO: this assumes that keystone always gives abs addr when disassembling, but it might not be true
        disassembled = self.p.disassembler.to_asm_string(disassembled)
        for test_addr in [addr - 0x10000, addr + 0x10000]:
            re_assembled = self.p.assembler.assemble(
                disassembled, test_addr, is_thumb=is_thumb
            )
            re_disassembled = self.p.disassembler.disassemble(
                re_assembled, test_addr, is_thumb=is_thumb
            )[0]
            re_disassembled = self.p.disassembler.to_asm_string(re_disassembled)
            if re_disassembled != disassembled:
                return False
        return True



class X86Info:
    nop_bytes = b"\x90"
    nop_size = 1
    jmp_asm = "jmp {dst}"
    jmp_size = 5
    alignment = 4
    bits = 32
    is_variable_length_isa = True
    instr_size = -1  # variable length
    call_asm = "call {dst}"
    pc_reg_names = ["eip"]
    save_context_asm = """
    pusha
    """
    restore_context_asm = """
    popa
    """
